/* winlog.c - SYSLOG Receiver for Windows
 * Copyright(C) 2002  Yuichi Kobayashi <kobayasy@kobayasy.com>
 *
 * This program comes with NO WARRANTY, to the extent permitted by law.
 * You may redistribute it under the terms of the GNU General Public License;
 * see the file named LICENSE.txt for details.
 */

#define TITLE\
 "winlog"
#define ABOUT\
 TITLE" Ver.0.1  -  SYSLOG Receiver for Windows"
#define RIGHT\
 "Copyright(C) 2002  Yuichi Kobayashi <kobayasy@kobayasy.com>"
#define LICENSE\
 "This program comes with NO WARRANTY, to the extent permitted by law.\n"\
 "You may redistribute it under the terms of the GNU General Public License;"

#include <winsock.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#define P      "-"
#define PHELP  P"help"
#define PPORT  P"port="
#define PTFORM P"tform="

#define LOG_PORT 514

#define NSTR 1024

static struct {
  unsigned short port;
  char tform[NSTR];
} g = {
  LOG_PORT,
  "[%m/%d/%y %H:%M:%S] "
};

static unsigned short swends(unsigned short n) {
  return (n << 8) | (n >> 8);
}

static int syslogd(unsigned short port, char *tform) {
  int status = -1;
  SOCKET logsoc;
  struct sockaddr_in logaddr;
  char logpac[NSTR];
  int len;
  time_t t;
  char buffer[NSTR];

  if((logsoc = socket(AF_INET, SOCK_DGRAM, 0)) == INVALID_SOCKET)
    status = 1;
  else {
    memset(&logaddr, 0, sizeof(logaddr));
    logaddr.sin_family = AF_INET;
    logaddr.sin_addr.s_addr = INADDR_ANY;
    logaddr.sin_port = swends(port);
    if(bind(logsoc, (struct sockaddr *)&logaddr, sizeof(logaddr))
       == SOCKET_ERROR ) {
      status = (WSAGetLastError() == WSAEADDRINUSE) ? 2 : 3;
      goto error;
    }
    while(!0) {
      if((len = recv(logsoc, logpac, NSTR, 0)) == SOCKET_ERROR) {
        status = 4;
        goto error;
      }
      logpac[len] = 0;
      time(&t);
      strftime(buffer, NSTR, tform, localtime(&t));
      fprintf(stdout, "%s%s\n", buffer, logpac), fflush(stdout);
    }
    status = 0;
   error:
    closesocket(logsoc);
  }
  return status;
}
static const char *syslogderr[] = {
  "no error",
  "make socket failure",                /* socket() == INVALID_SOCKET */
  "bind port in use",         /* bind() == SOCKET_ERROR:WSAEADDRINUSE */
  "bind port failure",                      /* bind() == SOCKET_ERROR */
  "receive packet failure",             /* recvfrom() == SOCKET_ERROR */
  NULL
};

int main(int argc, char *argv[]) {
  int status = -1;
  char *pname = *argv;
  WSADATA wsadata;
  size_t size;

  if(WSAStartup(MAKEWORD(1,1), &wsadata)) {
    status = 1;
    fprintf(stderr, "error %d: initialize winsock failure\n", status);
  }
  else {
    while(*++argv)
      if(!strcmp(*argv, PHELP)) {
        fprintf(stdout,
                "%s\n"
                "%s\n"
                "%s\n"
                "\n"
                "Usage: %s [option]...\n"
                "\n"
                " option:\n"
                "  %s%*sprint this message\n"
                "  %sport%*sset syslog port\n"
                "  %sformat%*sset time format\n"
                "\n"
                " current setting as:\n"
                "  %s%hu %s\x22%s\x22\n"
                "\n",
                ABOUT, RIGHT, LICENSE,
                pname,
                PHELP,  30 - (strlen(PHELP)  + 0), "",
                PPORT,  30 - (strlen(PPORT)  + 4), "",
                PTFORM, 30 - (strlen(PTFORM) + 6), "",
                PPORT, g.port,
                PTFORM, g.tform );
        status = 0;
        goto error;
      }
      else if(!strncmp(*argv, PPORT, size = strlen(PPORT)))
        g.port = strtoul(*argv + size, NULL, 10);
      else if(!strncmp(*argv, PTFORM, size = strlen(PTFORM)))
        strcpy(g.tform, *argv + size);
      else {
        status = 2;
        fprintf(stderr, "error %d: invalid parameter %s\n",
                        status, *argv );
        goto error;
      }
    if(status = syslogd(g.port, g.tform))
      fprintf(stderr, "error %d: %s\n",
                      (status > 0) ? status + 100 : status,
                      (status < 0) ? "unknown failure"
                                   : syslogderr[status] );
   error:
    WSACleanup();
  }
  return status;
}
