/* winsntpc.c - SNTP Client for Windows
 * Copyright(C) 2002  Yuichi Kobayashi <kobayasy@kobayasy.com>
 *
 * This program comes with NO WARRANTY, to the extent permitted by law.
 * You may redistribute it under the terms of the GNU General Public License;
 * see the file named LICENSE.txt for details.
 */

#define TITLE\
 "WinSNTPC"
#define ABOUT\
 TITLE" Ver.0.7  -  SNTP Client for Windows"
#define RIGHT\
 "Copyright(C) 2002  Yuichi Kobayashi <kobayasy@kobayasy.com>"
#define LICENSE\
 "This program comes with NO WARRANTY, to the extent permitted by law.\n"\
 "You may redistribute it under the terms of the GNU General Public License;"

#ifndef CONSOLE
#define CONSOLE 0
#endif

#include <windows.h>
#include <winsock.h>
#include <crtl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define P         "-"
#define PHELP     P"help"
#define PNTP      P"ntp"
#define PTIME     P"time"
#define PTIMEOUT  P"timeout="
#define PRETRY    P"retry="
#define PPERIODIC P"periodic="
#define PQUIET    P"quiet"
#define PVERBOCE  P"verboce"
#define PLOG      P"log="

#define NTP_PORT  123
#define TIME_PORT  37
#define LOG_PORT  514

#define NSTR 1024

typedef struct {
  struct {
    signed char precision;
    signed char poll;
    unsigned char stratum;
    unsigned char mode:3, vn:3, li:2;
  } control;
  signed long root_delay;
  unsigned long root_dispersion;
  char reference_identifier[4];
  unsigned long long reference_timestamp;
  unsigned long long originate_timestamp;
  unsigned long long receive_timestamp;
  unsigned long long transmit_timestamp;
} NTPPAC;

#define FPORT 0x01
static struct {
  char host[NSTR];
  unsigned short port;
  enum {Sntpc, Timec} protocol;
  unsigned long timeout;
  unsigned int retry;
  unsigned long periodic;
  enum {Standard, Quiet, Verboce} message;
  char loghost[NSTR];
  unsigned short logport;
  unsigned char flags;
} g = {
  "ntpserver", NTP_PORT,
  Sntpc,
  3, 0, 0,
  Standard,
  "localhost", LOG_PORT,
  0
};

static unsigned long long getsystemtime(void) {
  unsigned long long timestamp;

  GetSystemTimeAsFileTime((FILETIME *)&timestamp);
  timestamp -= (unsigned long long)109207*24*60*60*10000000;
  timestamp = ((timestamp << 8) / 10000000) << 24;
  return timestamp;
}
static int setsystemtime(unsigned long long timestamp) {
  SYSTEMTIME systemtime;

  timestamp = ((timestamp >> 24) * 10000000) >> 8;
  timestamp += (unsigned long long)109207*24*60*60*10000000;
  return FileTimeToSystemTime((FILETIME *)&timestamp, &systemtime) &&
         SetSystemTime(&systemtime);
}

static unsigned short swends(unsigned short n) {
  return (n << 8) | (n >> 8);
}
static unsigned long swendl(unsigned long n) {
  return ((unsigned long)swends(n) << 16) | swends(n >> 16);
}
static unsigned long long swendll(unsigned long long n) {
  return ((unsigned long long)swendl(n) << 32) | swendl(n >> 32);
}
static NTPPAC *swendntppac(NTPPAC *ntppac) {
  *(unsigned long *)&ntppac->control =
    swendl(*(unsigned long *)&ntppac->control);
  ntppac->root_delay = swendl(ntppac->root_delay);
  ntppac->root_dispersion = swendl(ntppac->root_dispersion);
  ntppac->reference_timestamp = swendll(ntppac->reference_timestamp);
  ntppac->originate_timestamp = swendll(ntppac->originate_timestamp);
  ntppac->receive_timestamp = swendll(ntppac->receive_timestamp);
  ntppac->transmit_timestamp = swendll(ntppac->transmit_timestamp);
  return ntppac;
}

static unsigned long getinaddr(char *name) {
  unsigned long inaddr;
  struct hostent *host;

  if(((inaddr = inet_addr(name)) == INADDR_NONE) &&
     (host = gethostbyname(name)) )
    inaddr = *(unsigned long *)host->h_addr;
  return inaddr;
}

static int sntpc(char *host, unsigned short port,
                 unsigned long timeout ) {
  int status = -1;
  unsigned long inaddr;
  SOCKET ntpsoc;
  struct sockaddr_in ntpaddr;
  NTPPAC ntppac;
  struct timeval tv;
  fd_set readfds;
  int addrlen;
  unsigned long long destination_timestamp;
  signed long long systemtimeoffset;

  if((inaddr = getinaddr(host)) == INADDR_NONE)
    status = 1;
  else if((ntpsoc = socket(AF_INET, SOCK_DGRAM, 0)) == INVALID_SOCKET)
    status = 2;
  else {
    memset(&ntpaddr, 0, sizeof(ntpaddr));
    ntpaddr.sin_family = AF_INET;
    ntpaddr.sin_addr.s_addr = inaddr;
    ntpaddr.sin_port = swends(port);
    memset(&ntppac, 0, sizeof(ntppac));
    ntppac.control.vn = 4, ntppac.control.mode = 3;
    ntppac.transmit_timestamp = getsystemtime();
    swendntppac(&ntppac);
    switch(sendto(ntpsoc, (char *)&ntppac, sizeof(ntppac),
                  0, (struct sockaddr *)&ntpaddr, sizeof(ntpaddr) )) {
     case SOCKET_ERROR:
      status = 3;
      goto error;
      break;
     case sizeof(ntppac):
      break;
     default:
      status = 4;
      goto error;
    }
    memset(&tv, 0, sizeof(tv));
    tv.tv_sec = timeout;
    FD_ZERO(&readfds);
    FD_SET(ntpsoc, &readfds);
    switch(select(FD_SETSIZE, &readfds, NULL, NULL, &tv)) {
     case SOCKET_ERROR:
      status = 5;
      goto error;
      break;
     case 0:
      status = 6;
      goto error;
      break;
    }
    switch(addrlen = sizeof(ntpaddr),
           recvfrom(ntpsoc, (char *)&ntppac, sizeof(ntppac),
                    0, (struct sockaddr *)&ntpaddr, &addrlen ) ) {
     case SOCKET_ERROR:
      status = 7;
      goto error;
      break;
     case sizeof(ntppac):
      break;
     default:
      status = 8;
      goto error;
    }
    swendntppac(&ntppac);
    destination_timestamp = getsystemtime();
    if((ntppac.control.li == 3) ||
       (ntppac.control.vn != 4) || (ntppac.control.mode != 4) ||
       !ntppac.control.stratum || (ntppac.control.stratum & 0xf0) ||
       !ntppac.originate_timestamp ||
       !ntppac.receive_timestamp ||
       !ntppac.transmit_timestamp ) {
      status = 9;
      goto error;
    }
    systemtimeoffset =
      (signed long long)((ntppac.receive_timestamp -
                          ntppac.originate_timestamp ) +
                         (ntppac.transmit_timestamp -
                          destination_timestamp ) ) >> 1;
    if(!setsystemtime(getsystemtime() + systemtimeoffset)) {
      status = 10;
      goto error;
    }
    status = 0;
   error:
    closesocket(ntpsoc);
  }
  return status;
}
static const char *sntpcerr[] = {
  "adjusted systemtime to server",
  "host name lookup failure",           /* getinaddr() == INADDR_NONE */
  "make socket failure",                /* socket() == INVALID_SOCKET */
  "send packet failure",                  /* sendto() == SOCKET_ERROR */
  "send packet incomplete",                        /* sendto() != len */
  "wait response failure",                /* select() == SOCKET_ERROR */
  "no response from servre",                         /* select() == 0 */
  "receive packet failure",             /* recvfrom() == SOCKET_ERROR */
  "receive packet incomplete",                   /* recvfrom() != len */
  "unspecified response from server",        /* ntppac is unspecified */
  "can't set systemtime",                         /* !setsystemtime() */
  NULL
};

static int timec(char *host, unsigned short port,
                 unsigned long timeout ) {
  int status = -1;
  unsigned long inaddr;
  SOCKET timesoc;
  struct sockaddr_in timeaddr;
  struct timeval tv;
  fd_set readfds;
  unsigned long timepac;

  if((inaddr = getinaddr(host)) == INADDR_NONE)
    status = 1;
  else if((timesoc = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET)
    status = 2;
  else {
    memset(&timeaddr, 0, sizeof(timeaddr));
    timeaddr.sin_family = AF_INET;
    timeaddr.sin_addr.s_addr = inaddr;
    timeaddr.sin_port = swends(port);
    if(connect(timesoc, (struct sockaddr *)&timeaddr, sizeof(timeaddr))
       == SOCKET_ERROR )
      status = 3;
    else {
      memset(&tv, 0, sizeof(tv));
      tv.tv_sec = timeout;
      FD_ZERO(&readfds);
      FD_SET(timesoc, &readfds);
      switch(select(FD_SETSIZE, &readfds, NULL, NULL, &tv)) {
       case SOCKET_ERROR:
        status = 4;
        goto error;
        break;
       case 0:
        status = 5;
        goto error;
        break;
      }
      switch(recv(timesoc, (char *)&timepac, sizeof(timepac), 0)) {
       case SOCKET_ERROR:
        status = 6;
        goto error;
        break;
       case sizeof(timepac):
        break;
       default:
        status = 7;
        goto error;
      }
      timepac = swendl(timepac);
      if(!setsystemtime((unsigned long long)timepac << 32)) {
        status = 8;
        goto error;
      }
      status = 0;
     error:
      shutdown(timesoc, 2);
    }
    closesocket(timesoc);
  }
  return status;
}
static const char *timecerr[] = {
  "adjusted systemtime to server",
  "host name lookup failure",           /* getinaddr() == INADDR_NONE */
  "make socket failure",                /* socket() == INVALID_SOCKET */
  "can't connect to server",             /* connect() == SOCKET_ERROR */
  "wait response failure",                /* select() == SOCKET_ERROR */
  "no response from servre",                         /* select() == 0 */
  "receive packet failure",                 /* recv() == SOCKET_ERROR */
  "receive packet incomplete",                       /* recv() != len */
  "can't set systemtime",                         /* !setsystemtime() */
  NULL
};

static int syslog(char *message, char *title,
                  char *host, unsigned short port ) {
  int status = -1;
  unsigned long inaddr;
  SOCKET logsoc;
  char buffer[NSTR], *s;
  struct sockaddr_in logaddr;
  int len, n;

  if((inaddr = getinaddr(host)) == INADDR_NONE)
    status = 1;
  else if((logsoc = socket(AF_INET, SOCK_DGRAM, 0)) == INVALID_SOCKET)
    status = 2;
  else {
    if(title)
      sprintf(buffer, "%s: %s", title, message), s = buffer;
    else
      s = message;
    memset(&logaddr, 0, sizeof(logaddr));
    logaddr.sin_family = AF_INET;
    logaddr.sin_addr.s_addr = inaddr;
    logaddr.sin_port = swends(port);
    if((len = sendto(logsoc, s, n = strlen(s),
                     0, (struct sockaddr *)&logaddr, sizeof(logaddr) ))
       == SOCKET_ERROR )
      status = 3;
    else if(len != n)
      status = 4;
    else
      status = 0;
    closesocket(logsoc);
  }
  return status;
}
static const char *syslogerr[] = {
  "output message to syslog",
  "host name lookup failure",           /* getinaddr() == INADDR_NONE */
  "make socket failure",                /* socket() == INVALID_SOCKET */
  "send packet failure",                  /* sendto() == SOCKET_ERROR */
  "send packet incomplete",                        /* sendto() != len */
  NULL
};

static void puts_info(char *message, char *title) {
#if CONSOLE
  fputs(message, stdout);
#else
  MessageBox(NULL, message, title, MB_OK);
#endif
}
static void puts_notice(char *message, char *title) {
#if CONSOLE
  if(title)
    fprintf(stdout, "%s: %s", title, message);
  else
    fputs(message, stdout);
#else
  MessageBox(NULL, message, title, MB_OK);
#endif
}
static void puts_error(char *message, char *title) {
#if CONSOLE
  if(title)
    fprintf(stderr, "%s: %s", title, message);
  else
    fputs(message, stderr);
#else
  MessageBox(NULL, message, title, MB_OK|MB_ICONEXCLAMATION);
#endif
}
int main(int argc, char *argv[]) {
  int status = -1;
  char *pname = *argv;
  WSADATA wsadata;
  char buffer[NSTR];
  size_t size;
  char *s;
  unsigned int retry;

  if(WSAStartup(MAKEWORD(1,1), &wsadata)) {
    status = 1;
    sprintf(buffer, "error %d: initialize winsock failure\n", status);
    puts_error(buffer, TITLE);
  }
  else {
    while(*++argv)
      if(!strcmp(*argv, PHELP)) {
        char buffer1[NSTR], buffer2[NSTR];

        if(g.retry)
          sprintf(buffer1, " %s%lu", PRETRY, g.retry);
        else
          *buffer1 = 0;
        if(g.periodic)
          sprintf(buffer2, " %s%lu", PPERIODIC, g.periodic);
        else
          *buffer2 = 0;
        sprintf(buffer,
                "%s\n"
                "%s\n"
                "%s\n"
                "\n"
                "Usage: %s host[:port] [option]...\n"
                "\n"
                " option:\n"
                "  %s%*sprint this message\n"
                "  %s%*sadjust to ntp server\n"
                "  %s%*sadjust to time server\n"
                "  %ssecond%*sset timeout time in second units\n"
                "  %scount%*senable retry, retry count\n"
                "  %shour%*senable periodic, "
                           "interval time in hour units\n"
                "  %s%*ssuppress all messages\n"
                "  %s%*sbe verboce\n"
                "  %shost[:port]%*sset syslog host [and port]\n"
                "\n"
                " current configuration:\n"
                "  %s:%hu%s%s %s%lu%s%s%s%s %s%s:%hu\n"
                "\n",
                ABOUT, RIGHT, LICENSE,
                pname,
                PHELP,     30 - (strlen(PHELP)     +  0), "",
                PNTP,      30 - (strlen(PNTP)      +  0), "",
                PTIME,     30 - (strlen(PTIME)     +  0), "",
                PTIMEOUT,  30 - (strlen(PTIMEOUT)  +  6), "",
                PRETRY,    30 - (strlen(PRETRY)    +  5), "",
                PPERIODIC, 30 - (strlen(PPERIODIC) +  4), "",
                PQUIET,    30 - (strlen(PQUIET)    +  0), "",
                PVERBOCE,  30 - (strlen(PVERBOCE)  +  0), "",
                PLOG,      30 - (strlen(PLOG)      + 11), "",
                g.host, g.port,
                (g.protocol == Sntpc) ? " "PNTP : "",
                (g.protocol == Timec) ? " "PTIME : "",
                PTIMEOUT, g.timeout,
                buffer1,
                buffer2,
                (g.message == Quiet)   ? " "PQUIET   : "",
                (g.message == Verboce) ? " "PVERBOCE : "",
                PLOG, g.loghost, g.logport );
        puts_info(buffer, TITLE);
        status = 0;
        goto error;
      }
      else if(!strcmp(*argv, PNTP)) {
        g.protocol = Sntpc;
        if(!(g.flags & FPORT))
          g.port = NTP_PORT;
      }
      else if(!strcmp(*argv, PTIME)) {
        g.protocol = Timec;
        if(!(g.flags & FPORT))
          g.port = TIME_PORT;
      }
      else if(!strncmp(*argv, PTIMEOUT, size = strlen(PTIMEOUT)))
        g.timeout = strtoul(*argv + size, NULL, 10);
      else if(!strncmp(*argv, PRETRY, size = strlen(PRETRY)))
        g.retry = strtoul(*argv + size, NULL, 10);
      else if(!strncmp(*argv, PPERIODIC, size = strlen(PPERIODIC)))
        g.periodic = strtoul(*argv + size, NULL, 10);
      else if(!strcmp(*argv, PQUIET))
        g.message = Quiet;
      else if(!strcmp(*argv, PVERBOCE))
        g.message = Verboce;
      else if(!strncmp(*argv, PLOG, size = strlen(PLOG))) {
        strcpy(g.loghost, *argv + size);
        if(s = strchr(g.loghost, ':')) {
          *s++ = 0;
          g.logport = strtoul(s, NULL, 10);
        }
      }
      else if(strncmp(*argv, P, strlen(P))) {
        strcpy(g.host, *argv);
        if(s = strchr(g.host, ':')) {
          *s++ = 0;
          g.port = strtoul(s, NULL, 10);
          g.flags |= FPORT;
        }
      }
      else {
        status = 2;
        sprintf(buffer, "error %d: invalid parameter %s\n",
                        status, *argv );
        puts_error(buffer, TITLE);
        goto error;
      }
    do {
      retry = 0;
      do {
        switch(g.protocol) {
         case Sntpc:
          status = sntpc(g.host, g.port, g.timeout);
          s = (status < 0) ? "unknown failure"
                           : (char *)sntpcerr[status];
          if(status > 0)
            status += 100;
          break;
         case Timec:
          status = timec(g.host, g.port, g.timeout);
          s = (status < 0) ? "unknown failure"
                           : (char *)timecerr[status];
          if(status > 0)
            status += 200;
          break;
         default:
          status = 0;
          s = NULL;
        }
        if(s) {
          if(status)
            sprintf(buffer, retry ? "error %d: %s (retry %u)"
                                  : "error %d: %s",
                            status, s, retry );
          else
            sprintf(buffer, retry ? "%s (retry %u)" : "%s", s, retry);
          syslog(buffer, TITLE, g.loghost, g.logport);
          s = buffer;
        }
      } while(status && (retry++ < g.retry));
      if(s) {
        strcat(s, "\n");
        switch(g.message) {
         case Standard:
          if(status)
            puts_error(s, TITLE);
          break;
         case Quiet:
          break;
         case Verboce:
          if(status)
            puts_error(s, TITLE);
          else
            puts_notice(s, TITLE);
          break;
        }
      }
      if(g.periodic)
        Sleep(g.periodic * (unsigned long)60*60*1000);
    } while(g.periodic);
   error:
    WSACleanup();
  }
  return status;
}
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                   LPSTR lpszCmdLine, int nCmdShow ) {
  return main(__argc, __argv);
}
